# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os.path
from pathlib import Path


def get_version() -> str:
    parent_folder_path = Path(__file__).parent.absolute()
    with open(os.path.join(parent_folder_path, "VERSION")) as f:
        return f.read().rstrip()
